#ifndef HASHRAT_FRONTEND_H
#define HASHRAT_FRONTEND_H

#include "common.h"

//provides certain values for use in frontends like .cgi and xdialog

extern const char *EncodingNames[];
extern const char *EncodingDescriptions[];
extern int Encodings[];

extern const char *LineEndingNames[];
extern const char *LineEndingDescriptions[];

#endif
