/*
Copyright (c) 2025 Colum Paget <colums.projects@googlemail.com>
* SPDX-License-Identifier: GPL-3.0
*/

#ifndef LIBUSEFUL_SECCOMP_H
#define LIBUSEFUL_SECCOMP_H

#include "includes.h"

int SeccompAddRules(const char *RuleList);

#endif
