#include "frontend.h"

const char *LineEndingNames[]= {"none","lf","crlf","cr",NULL};
const char *LineEndingDescriptions[]= {"None","Unix newline","DOS carriage-return and newline","carriage-return",NULL};

