/*
Copyright (c) 2015 Colum Paget <colums.projects@googlemail.com>
* SPDX-License-Identifier: GPL-3.0
*/

#ifndef LIBUSEFUL_CONTENT_TYPE_H
#define LIBUSEFUL_CONTENT_TYPE_H

#include "includes.h"

const char *ContentTypeForFile(const char *File);

#endif
