/*
Copyright (c) 2015 Colum Paget <colums.projects@googlemail.com>
* SPDX-License-Identifier: GPL-3.0
*/

#ifndef LIBUSEFUL_HASH_MD5_H
#define LIBUSEFUL_HASH_MD5_H



#ifdef __cplusplus
extern "C" {
#endif


#include "Hash.h"

void HashRegisterMD5();

#ifdef __cplusplus
}
#endif



#endif
